/*
 * Copyright (c) 2015 N. Billard.
 */
package com.massa.mrules.demos.advanced;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;

import org.junit.Assert;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameters;

import com.massa.log.Log;
import com.massa.log.LogFactory;
import com.massa.mrules.context.compile.CompilationLevel;
import com.massa.mrules.context.execute.MExecutionContext;
import com.massa.mrules.demos.advanced.model.Entity;
import com.massa.mrules.demos.advanced.model.ExtendedEntity;
import com.massa.mrules.demos.advanced.model.SearchEntity;
import com.massa.mrules.exception.MConfigurationException;
import com.massa.mrules.exception.MExecutionException;
import com.massa.mrules.factory.ser.SerializationFactory;
import com.massa.mrules.factory.xml.RichXMLFactory;
import com.massa.mrules.set.IMruleExecutionSet;

@RunWith(Parameterized.class)
public class AdvancedTest {
	private static final Log LOG = LogFactory.getLog(AdvancedTest.class.getName());

	private static final Object[][] round = new Object [][] {
		{"15.5687", 15.57D},
		{"12.5", 12.5D},
		{new BigDecimal("18.441"), 18.44D},
		{new BigDecimal("18.445"), 18.44D},
		{new BigDecimal("18.455"), 18.46D},
		{new BigDecimal("18.458"), 18.46D},
		{5.3265D, 5.33D},
		{6.1457F, 6.15D},
		{8, 8D},
		{9L, 9D},
	};

	final CompilationLevel level;

	public AdvancedTest(final CompilationLevel level) {
		this.level = level;
	}

	@Parameters
	public static Collection<Object[]> configs() {
		return Arrays.asList(new Object[][] {
			{CompilationLevel.STANDARD},
			{CompilationLevel.LIGHT}
		});
	}

	private static final String XML = "com/massa/mrules/demos/advanced/rules.xml";

	@Test
	public void advancedDemo() throws MConfigurationException, MExecutionException, IOException {
		final RichXMLFactory factory = new RichXMLFactory(this.level);
		final InputStream xml = ClassLoader.getSystemResourceAsStream(AdvancedTest.XML);
		final IMruleExecutionSet eng = factory.read(xml);

		@SuppressWarnings("unchecked")
		final MExecutionContext<Entity> ctxt = (MExecutionContext<Entity>) eng.getContextFactory().getExecutionContext();

		final Entity entity = this.prepareEntity(new Entity());
		ctxt.setInput(entity);

		this.reset(entity);
		ctxt.execute();

		final ExtendedEntity extentity = this.prepareExtEntity();
		ctxt.setInput(extentity);

		this.reset(extentity);
		ctxt.execute();

		Assert.assertNotNull(entity.getSearchValues());
		Assert.assertEquals(3, entity.getSearchValues().size());
		Assert.assertEquals("VALUE2", entity.getSearchValues().get(0));
		Assert.assertEquals("NULL", entity.getSearchValues().get(1));
		Assert.assertNull(entity.getSearchValues().get(2));

		Assert.assertNotNull(entity.getNumberResult());
		Assert.assertEquals(AdvancedTest.round.length, entity.getNumberResult().size());
		for (int i = 0 ; i < AdvancedTest.round.length ; i++) {
			Assert.assertEquals(AdvancedTest.round[i][1], entity.getNumberResult().get(i));
		}

		Assert.assertNotNull(entity.getMyStringResults());
		Assert.assertEquals("AZE", entity.getMyStringResults()[0]);
		Assert.assertEquals("N/A", entity.getMyStringResults()[1]);

		//Time check
		AdvancedTest.LOG.info("Beginning performance test.");
		ctxt.setLoggerName(".perf"); //to avoid debug and info.
		ctxt.setOut(null); //To avoid thousands of printings.
		entity.setSysCommand(null); //System calls are very time consuming
		extentity.setSysCommand(null); //System calls are very time consuming
		final long t1 = System.currentTimeMillis();
		for (int i = 0 ; i < 10000 ; i++) {
			this.reset(entity);
			ctxt.setInput(entity);
			ctxt.execute();
			this.reset(extentity);
			ctxt.setInput(extentity);
			ctxt.execute();
		}
		AdvancedTest.LOG.info("Time for 10000 tests : " + (System.currentTimeMillis() - t1));

		//Read / Write tests.
		AdvancedTest.LOG.info("Human readable version:\n" + eng.toString());

		final StringWriter w = new StringWriter();
		factory.write(w, eng);
		AdvancedTest.LOG.info(w.toString());

		final ByteArrayOutputStream o = new ByteArrayOutputStream();
		final SerializationFactory sfactory = new SerializationFactory(this.level);
		sfactory.write(o, eng);

		Assert.assertEquals(eng, eng.clone());
		Assert.assertEquals(eng, factory.read(new StringReader(w.toString())));
		Assert.assertEquals(eng, sfactory.read(new ByteArrayInputStream(o.toByteArray())));
	}

	private ExtendedEntity prepareExtEntity() {
		final ExtendedEntity entity = new ExtendedEntity();
		this.prepareEntity(entity);
		return entity;
	}

	private Entity prepareEntity(final Entity entity) {
		entity.setSearchList(new ArrayList<SearchEntity>());
		entity.getSearchList().add(new SearchEntity(null, "NULL"));
		entity.getSearchList().add(new SearchEntity("CRIT1", "VALUE1"));
		entity.getSearchList().add(new SearchEntity("CRIT2", "VALUE2"));
		entity.getSearchList().add(new SearchEntity("CRIT3", "VALUE3"));

		entity.setNumberList(new ArrayList<Object>());
		for (final Object[] nums: AdvancedTest.round) {
			entity.getNumberList().add(nums[0]);
		}

		entity.setSysCommand("cmd /C echo My command result.");

		entity.setMyString("AZERTY");

		return entity;
	}

	private void reset(final Entity entity) {
		entity.setSearchValues(null);
		entity.setNumberResult(null);
	}
}
